"""
Module for constructing GUI components:
    -   Main window.
    -   Dialog for selecting a file.
    -   Dialog for saving a file.
"""

import os
import string
import platform
from pathlib import Path
from typing import Optional, List
from nicegui import ui
from nicegui.events import GenericEventArguments
import broh5.lib.utilities as util


# ============================================================================
#                   Global parameters for the GUI
# ============================================================================

MARKER_LIST = [",", ".", "o", "x"]
CMAP_LIST = ["gray", "inferno", "afmhot", "viridis", "magma"]
AXIS_LIST = [0, 1, 2]
FONT_STYLE = "font-size: 105%; font-weight: bold"
DISPLAY_TYPE = ["plot", "table"]
UPDATE_RATE = 0.2  # second
RATIO = 0.65  # Ratio for adjusting size between image/plot and screen
MAX_FIG_SIZE = [12.0, 9.0]
MAX_PLOT_SIZE = [9.0, 7.0]
INPUT_EXT = ["hdf", "nxs", "h5", "hdf5"]
PLOT_COLOR = "blue"
HEADER_COLOR = "#3874c8"
HEADER_TITLE = "BROWSER-BASED HDF VIEWER"
LEFT_DRAWER_COLOR = "#d7e3f4"
TREE_BGR_COLOR = "#f8f8ff"
BOX_LINE_COLOR = "lime"
BOX_LINE_WIDTH = 3


class GuiRendering:
    """
    A class to build the graphical user interface for an HDF viewer.

    This class creates various UI elements like headers, buttons,
    sliders, tables, and plots to facilitate user interaction with HDF data.

    Attributes
    ----------
    fig_size : tuple
        Dimensions for the figure in the UI.
    plot_size : tuple
        Dimensions for the histogram plot in the UI.
    tree_container : UI column
        Container for the HDF tree structure.
    select_file_button : UI button
        Button to trigger file selection.
    file_path_display : UI label
        Label to display selected file path.
    hdf_key_display : UI label
        Label to display HDF key.
    hdf_value_display : UI label
        Label to display HDF value.
    axis_list : UI select
        Dropdown to select axis for slicing.
    cmap_list : UI select
        Dropdown to select color map for plots.
    enable_zoom : UI checkbox
        Check-box to enable/disable image zoom.
    zoom_list : UI select
        Dropdown to select zooming ratio.
    enable_profile :  UI checkbox
        Check-box to enable/disable the intensity-profile plot
    profile_list : UI select
        Dropdown to select the direction of the intensity profile.
    save_image_button : UI button
        Button to save current image.
    display_type : UI select
        Dropdown to choose the display type (plot or table).
    marker_list : UI select
        Dropdown to select marker type for plots.
    save_data_button : UI button
        Button to save current data.
    main_slider : UI slider
        Slider to navigate through slices of 3D data.
    main_table : UI table
        Table to display data in tabular form.
    main_plot : UI matplotlib
        Matplotlib element to display plots.
    zoom_profile_plot : UI pyplot
        Pyplot element to display zooming of an image or intensity-profile.
    min_slider : UI slider
        Slider to adjust the minimum value for image contrast.
    max_slider : UI slider
        Slider to adjust the maximum value for image contrast.
    reset_button : UI button
        Button to reset adjustments.
    histogram_plot : UI pyplot
        Pyplot element to display histogram of an image.
    image_info_table : UI table
        Table to display statistical information of an image.

    Methods
    -------
    init_gui()
        Initializes and constructs the GUI elements.
    """
    def __init__(self):
        super().__init__()
        # Initial parameters
        (sc_height, sc_width, dpi) = util.get_height_width_screen()
        hei_size = RATIO * sc_width / dpi
        wid_size = RATIO * sc_height / dpi
        self.dpi = dpi
        self.fig_size = (min(hei_size, MAX_FIG_SIZE[0]),
                         min(wid_size, MAX_FIG_SIZE[1]))
        self.plot_size = (min(hei_size, MAX_PLOT_SIZE[0]),
                          min(wid_size, MAX_PLOT_SIZE[1]))
        self.tree_container = None
        self.select_file_button = None
        self.file_path_display = None
        self.hdf_key_display = None
        self.hdf_value_display = None
        self.axis_list = None
        self.cmap_list = None
        self.enable_zoom = None
        self.zoom_list = None
        self.enable_profile = None
        self.profile_list = None
        self.save_image_button = None
        self.display_type = None
        self.marker_list = None
        self.save_data_button = None
        self.main_slider = None
        self.main_table = None
        self.main_plot = None
        self.zoom_profile_plot = None
        self.min_slider = None
        self.max_slider = None
        self.reset_button = None
        self.histogram_plot = None
        self.image_info_table = None
        self.tab_one = None
        self.tab_two = None
        self.panel_tabs = None
        self.init_gui()

    def init_gui(self):
        """
        Initializes and constructs the various elements of the GUI.

        This method sets up headers, drawers, rows, labels, buttons, sliders,
        tables, and plots to create an interactive user interface.
        """
        # For the header
        with ui.header().style("background-color: " + HEADER_COLOR).classes(
                "items-center justify-between"):
            ui.label(HEADER_TITLE).style(FONT_STYLE)

        # For the left drawer, used to display a hdf tree.
        with ui.left_drawer(fixed=True, bottom_corner=True).style(
                "background-color: " + LEFT_DRAWER_COLOR):
            with ui.row():
                self.tree_container = ui.column()
                with self.tree_container:
                    self.select_file_button = ui.button(
                        "Select file").props("icon=folder")

        # Layout for the main page.
        with ui.column().classes("w-full no-wrap gap-1"):
            # For displaying file-path, key, and value of a hdf/nxs/h5 file
            with ui.row().classes("w-full no-wrap"):
                with ui.row().classes("w-1/3 items-center"):
                    ui.label("File path: ").style(FONT_STYLE)
                    self.file_path_display = ui.label("")
                with ui.row().classes("w-1/3 items-center"):
                    ui.label("Key: ").style(FONT_STYLE)
                    self.hdf_key_display = ui.label("")
                with ui.row().classes("w-1/3 items-center"):
                    ui.label("Value: ").style(FONT_STYLE)
                    self.hdf_value_display = ui.label("")
            ui.separator()

            # For ui-components used to interact with data.
            with ui.row().classes("w-full justify-between items-center"):
                # For ui-components used to interact with 3d data.
                with ui.row().classes("items-center"):
                    ui.label("Axis: ").style(FONT_STYLE)
                    self.axis_list = ui.select(AXIS_LIST, value=AXIS_LIST[0])
                with ui.row().classes("items-center"):
                    ui.label("Color map: ").style(FONT_STYLE)
                    self.cmap_list = ui.select(CMAP_LIST, value=CMAP_LIST[0])
                with ui.row().classes("items-center"):
                    self.enable_zoom = ui.checkbox('Zoom')
                    self.zoom_list = ui.select(['2x', '4x', '8x'], value="2x")
                with ui.row().classes("items-center"):
                    self.enable_profile = ui.checkbox('Profile')
                    self.profile_list = ui.select(['vertical', 'horizontal'],
                                                  value='horizontal')
                self.save_image_button = ui.button("Save image")

                # For ui-components used to interact with 1d/2d data.
                with ui.row().classes("items-center"):
                    ui.label("Display: ").style(FONT_STYLE)
                    self.display_type = ui.select(DISPLAY_TYPE,
                                                  value=DISPLAY_TYPE[0])
                with ui.row().classes("items-center"):
                    ui.label("Marker: ").style(FONT_STYLE)
                    self.marker_list = ui.select(MARKER_LIST,
                                                 value=MARKER_LIST[0])
                self.save_data_button = ui.button("Save data")

            # Slider for slicing 3d dataset
            with ui.row().classes("w-full items-center no-wrap"):
                ui.label("Slice: ").style(FONT_STYLE)
                self.main_slider = ui.slider(min=0, max=100, value=0).props(
                    "label-always").on("update:model-value",
                                       throttle=UPDATE_RATE,
                                       leading_events=False)

            # Tabs for data visualization and displaying image information
            tabs = ui.tabs().classes('w-full')
            with tabs:
                self.tab_one = ui.tab('Data visualization').style(
                    "background-color: " + TREE_BGR_COLOR)
                self.tab_two = ui.tab('Image information').style(
                    "background-color: " + TREE_BGR_COLOR)
            self.panel_tabs = ui.tab_panels(tabs, value=self.tab_one).classes(
                'w-full')
            with self.panel_tabs:
                # Tab 1 for data visualization
                with ui.tab_panel(self.tab_one):
                    # For display data as an image, table, or plot
                    self.main_table = ui.table(columns=[], rows=[],
                                               row_key="Index")
                    with ui.row().classes("w-full justify-left items-center"):
                        self.main_plot = ui.matplotlib(figsize=self.fig_size,
                                                       dpi=self.dpi)
                        self.zoom_profile_plot = ui.matplotlib(
                            figsize=self.fig_size, dpi=self.dpi)

                    # Sliders for adjust the contrast of an image.
                    with ui.row().classes(
                            "w-full justify-between no-wrap items-center"):
                        ui.label("Min: ").style(FONT_STYLE)
                        self.min_slider = ui.slider(min=0, max=254,
                                                    value=0).props(
                            "label-always").on("update:model-value",
                                               throttle=UPDATE_RATE,
                                               leading_events=False)

                        ui.label("Max: ").style(FONT_STYLE)
                        self.max_slider = ui.slider(min=1, max=255,
                                                    value=255).props(
                            "label-always").on("update:model-value",
                                               throttle=UPDATE_RATE,
                                               leading_events=False)
                        self.reset_button = ui.button("Reset")
                # Tab 2 for showing image information
                with ui.tab_panel(self.tab_two):
                    with ui.row().classes(
                            "w-full justify-between no-wrap items-center"):
                        self.histogram_plot = ui.pyplot(figsize=self.plot_size,
                                                        close=False,
                                                        ).classes("w-full")
                        self.image_info_table = ui.table(columns=[],
                                                         rows=[],
                                                         row_key="name")


class FilePicker(ui.dialog):
    """
    A dialog class for file picking in the GUI. Codes are adapted from an
    example of NiceGUI:
    https://github.com/zauberzeug/nicegui/tree/main/examples/local_file_picker

    Allows users to browse and select files from the local filesystem where
    the application is running.

    Parameters
    ----------
    directory : str
        The starting directory for the file picker.
    upper_limit : str, optional
        The upper directory limit for browsing. None by default.
    show_hidden_files : bool, optional
        Flag to show hidden files. False by default.
    allowed_extensions : List of str, optional
        List of allowed file extensions for filtering. None by default.

    Methods
    -------
    check_extension(filename: str)
        Check if the given filename has an allowed extension.
    add_drives_toggle()
        Add a toggle for drive selection on Windows systems.
    update_grid()
        Update the file grid based on the current directory and filters.
    handle_double_click(GenericEventArguments)
        Handle double click events on the file grid.
    handle_ok()
        Handle the OK button, click to submit the selected file path.
    """
    def __init__(self, directory: str, *,
                 upper_limit: Optional[str] = None,
                 show_hidden_files: bool = False,
                 allowed_extensions: Optional[List[str]] = None) -> None:
        super().__init__()
        self.show_hidden_files = show_hidden_files
        self.allowed_extensions = allowed_extensions
        self.drives_toggle = None
        self.path = Path(directory).expanduser()
        if upper_limit is None:
            self.upper_limit = None
        else:
            self.upper_limit = Path(
                directory if upper_limit is ... else upper_limit).expanduser()
        with self, ui.card():
            self.add_drives_toggle()
            self.grid = ui.aggrid(
                {'columnDefs': [{'field': 'name', 'headerName': 'File'}],
                 'rowSelection': 'single'}, html_columns=[0]).classes(
                'w-96').on('cellDoubleClicked', self.handle_double_click)
            with ui.row().classes('w-full justify-end'):
                ui.button('Cancel', on_click=self.close).props('outline')
                ui.button('Ok', on_click=self.handle_ok)
        self.update_grid()

    def check_extension(self, filename: str) -> bool:
        """Check if the filename has an allowed extension."""
        if self.allowed_extensions is None:
            return True
        else:
            return filename.split('.')[-1].lower() in self.allowed_extensions

    def add_drives_toggle(self):
        """Give a list of available drivers in a WinOS computer"""
        if platform.system() == 'Windows':
            drives = ['%s:\\' % d for d in string.ascii_uppercase if
                      os.path.exists('%s:' % d)]
            if self.path != "" or self.path != ".":
                select_drive = os.path.splitdrive(self.path)[0] + "\\"
            else:
                self.path = Path(drives[0]).expanduser()
                select_drive = drives[0]
            self.drives_toggle = ui.toggle(drives, value=select_drive,
                                           on_change=self.__update_drive)

    def __update_drive(self):
        if self.drives_toggle:
            self.path = Path(self.drives_toggle.value).expanduser()
            self.update_grid()

    def update_grid(self) -> None:
        paths = list(self.path.glob('*'))
        if not self.show_hidden_files:
            paths = [p for p in paths if not p.name.startswith('.')]
        if self.allowed_extensions:
            paths = [p for p in paths if
                     p.is_dir() or self.check_extension(p.name)]
        paths.sort(key=lambda p: p.name.lower())
        paths.sort(key=lambda p: not p.is_dir())

        self.grid.options['rowData'] = [
            {'name': f'📁 <strong>{p.name}</strong>' if p.is_dir() else p.name,
             'path': str(p), } for p in paths]
        if (self.upper_limit is None
                and self.path != self.path.parent
                or self.upper_limit is not None
                and self.path != self.upper_limit):
            self.grid.options['rowData'].insert(0, {
                'name': '📁 <strong>..</strong>',
                'path': str(self.path.parent), })
        self.grid.update()

    def handle_double_click(self, e: GenericEventArguments) -> None:
        self.path = Path(e.args['data']['path'])
        if self.path.is_dir():
            self.update_grid()
        else:
            if self.path:
                self.submit(str(self.path))
            else:
                return

    async def handle_ok(self):
        try:
            self.update_grid()
            rows = await self.grid.get_selected_rows()
            if rows:
                fpath = [r['path'] for r in rows]
                if fpath:
                    selected_path = Path(fpath[0])
                    if selected_path.suffix in {'.h5', '.hdf', '.nxs'}:
                        self.submit(str(selected_path))
                    else:
                        ui.notify("Please select a file with the extension "
                                  ".h5, .hdf, or .nxs!")
                        return
                else:
                    ui.notify("No file path found in the selected rows")
                    return
            else:
                ui.notify("No rows selected. Try double-clicking instead!")
                return
        except Exception as e:
            ui.notify(f"An error occurred: {e}")
            return


class FileSaver(ui.dialog):
    """
    A dialog class for saving files in the GUI.

    Allows users to specify a file name and directory for saving files.

    Parameters
    ----------
    directory : str
        Starting directory for the file saver.
    upper_limit : str, optional
        Upper directory limit for browsing. None by default.
    show_hidden_files : bool, optional
        Flag to show hidden files. False by default.
    title : str, optional
        Title for the file-name input-field. 'File name' by default.

    Methods
    -------
    add_drives_toggle()
        Add a toggle for drive selection on Windows systems.
    update_grid() -> None
        Update the file grid based on the current directory.
    handle_double_click(e: GenericEventArguments) -> None
        Handle double-click events on the file grid.
    handle_save()
        Handle the Save button, click to submit the specified file path.
    create_folder_dialog()
        Open a dialog for creating a new folder.
    create_folder(folder_name: str, dialog: ui.dialog)
        Create a new folder with the specified name.
    """
    def __init__(self, directory: str, *, upper_limit: Optional[str] = None,
                 show_hidden_files: bool = False,
                 title: Optional[str] = 'File name') -> None:
        super().__init__()
        self.show_hidden_files = show_hidden_files
        self.drives_toggle = None
        self.path = Path(directory).expanduser()
        self.title = title
        if upper_limit is None:
            self.upper_limit = None
        else:
            self.upper_limit = Path(
                directory if upper_limit is ... else upper_limit).expanduser()

        with self, ui.card():
            self.add_drives_toggle()
            self.grid = ui.aggrid(
                {'columnDefs': [{'field': 'name', 'headerName': 'File'}],
                 'rowSelection': 'single'}, html_columns=[0]).classes(
                'w-96').on('cellDoubleClicked', self.handle_double_click)
            # Input field for filename
            self.filename_input = ui.input(self.title).classes(
                'w-full justify-between').on('keydown.enter',
                                             self.handle_save)
            with ui.row().classes('w-full justify-between'):
                ui.button('Create Folder',
                          on_click=self.create_folder_dialog).props('outline')
                ui.button('Cancel', on_click=self.close).props('outline')
                ui.button('Save', on_click=self.handle_save)
        self.update_grid()

    def add_drives_toggle(self):
        """Give a list of available drivers in a WinOS computer"""
        if platform.system() == 'Windows':
            drives = ['%s:\\' % d for d in string.ascii_uppercase if
                      os.path.exists('%s:' % d)]
            if self.path != "" or self.path != ".":
                select_drive = os.path.splitdrive(self.path)[0] + "\\"
            else:
                self.path = Path(drives[0]).expanduser()
                select_drive = drives[0]
            self.drives_toggle = ui.toggle(drives, value=select_drive,
                                           on_change=self.__update_drive)

    def __update_drive(self):
        if self.drives_toggle:
            self.path = Path(self.drives_toggle.value).expanduser()
            self.update_grid()

    def update_grid(self) -> None:
        paths = list(self.path.glob('*'))
        if not self.show_hidden_files:
            paths = [p for p in paths if not p.name.startswith('.')]
        paths.sort(key=lambda p: p.name.lower())
        paths.sort(key=lambda p: not p.is_dir())

        self.grid.options['rowData'] = [
            {'name': f'📁 <strong>{p.name}</strong>' if p.is_dir() else p.name,
             'path': str(p)} for p in paths]
        if (self.upper_limit is None
                and self.path != self.path.parent
                or self.upper_limit is not None
                and self.path != self.upper_limit):
            self.grid.options['rowData'].insert(0, {
                'name': '📁 <strong>..</strong>',
                'path': str(self.path.parent)})
        self.grid.update()

    def handle_double_click(self, e: GenericEventArguments) -> None:
        self.path = Path(e.args['data']['path'])
        if self.path.is_dir():
            self.update_grid()
        else:
            self.filename_input.value = self.path.name
            self.path = self.path.parent

    def handle_save(self):
        filename = self.filename_input.value
        if not filename:
            ui.notify('File name cannot be empty!')
            return
        save_path = self.path / filename
        save_path_str = str(save_path).replace('\\', '/')
        self.submit(save_path_str)

    async def create_folder_dialog(self):
        """Open a dialog to get the name of the new folder and create it."""
        with ui.dialog().classes('w-100 h-100') as dialog, ui.card():
            with ui.column():
                folder_name_input = ui.input('Folder Name').classes(
                    'w-full justify-between')
                with ui.row():
                    ui.button('Cancel', on_click=dialog.close).props('outline')
                    ui.button('Create', on_click=lambda: self.create_folder(
                        folder_name_input.value, dialog))
        await dialog

    def create_folder(self, folder_name: str, dialog: ui.dialog):
        if not folder_name:
            ui.notify('Folder name cannot be empty!')
            return
        new_folder_path = self.path / folder_name
        if new_folder_path.exists():
            ui.notify(f"A folder named '{folder_name}' already exists!")
            return
        new_folder_path.mkdir(parents=True, exist_ok=True)
        self.update_grid()
        dialog.close()



import os
import sys
import signal
import socket
import argparse
from nicegui import ui, app
from broh5.lib.interactions import GuiInteraction
from broh5 import __version__

display_msg = """
===============================================================================

Web-browser-based GUI (Graphical User Interface) software for viewing HDF files

===============================================================================

Type: broh5 to run the software
Exit the software by pressing: Ctrl + C

===============================================================================
"""


def handle_shutdown(main_app):
    """
    Handle the shutdown process of the application.
    """
    if main_app is not None:
        main_app.shutdown()
    print("\n===============")
    print(" Exit the app!")
    print("===============\n")


def signal_handler(*args):
    """
    Handle the signal for graceful shutdown of the application.

    Parameters
    ----------
    *args
        Variable length argument list.
    """
    sys.exit(0)


def check_port(port):
    """
    Check if a given port is available.

    Parameters
    ----------
    port : int
        The port number to check.

    Returns
    -------
    bool
        True if the port is already in use, False otherwise.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0


def parse_args():
    """
    Parse command-line arguments.

    Returns
    -------
    argparse.Namespace
        Parsed arguments.
    """
    parser = argparse.ArgumentParser(description=display_msg,
                                     formatter_class=
                                     argparse.RawDescriptionHelpFormatter)
    parser.add_argument("--version", action="version",
                        version=f"%(prog)s {__version__}")
    parser.add_argument("--port", type=int, default=8180,
                        help="Specify the port to run broh5 on "
                             "(default: 8180)")
    args = parser.parse_args()
    return args


def main():
    """
    Main function to start the application.
    """
    args = parse_args()
    if check_port(args.port):
        print("\n!!! Port {} is already in use. Please specify a "
              "different port using --port !!!\n".format(args.port))
        sys.exit(1)
    signal.signal(signal.SIGINT, signal_handler)  # Back-up shutdown
    try:
        broh5_app = None

        @ui.page('/')
        def main_page():
            global broh5_app
            broh5_app = GuiInteraction()

        app.on_shutdown(lambda: handle_shutdown(broh5_app))
        os.environ["NO_NETIFACES"] = "True"
        app.on_startup(
            lambda: print("Access Broh5 at urls: {}".format(app.urls.union())))
        ui.run(reload=False, title="Browser-based Hdf Viewer", port=args.port,
               show_welcome_message=False)
    except Exception as error:
        print(f"An error occurred: {error}")
        sys.exit(0)


if __name__ in {"__main__", "__mp_main__"}:
    main()



"""
This module links user interactions to the responses of the Broh5 software.
"""

import os
import h5py
import hdf5plugin
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.lines import Line2D
from nicegui import ui, events
import broh5.lib.rendering as re
import broh5.lib.utilities as util
from broh5.lib.rendering import GuiRendering, FilePicker, FileSaver


class GuiInteraction(GuiRendering):
    """
    A subclass of GuiRendering that provides specific functionalities for
    interacting with the GUI elements in Broh5.

    This class handles user actions such as file selection, branch/leaf
    selection in an HDF tree, image saving, or data display; resets and
    updates the GUI in response to these interactions.

    Attributes
    ----------
    current_state : tuple or None
        Current state of the GUI: file path, HDF key, slider values, ...
    columns : list or None
        Columns for displaying data in a table.
    rows : list or None
        Rows for displaying data in a table.
    image : np.ndarray or None
        Current slice from a 3D dataset.
    current_slice : tuple or None
        Information about the current slice being displayed.
    data_1d_2d : np.ndarray or None
        Current 1D or 2D data being displayed.
    timer : UI object
        To update the GUI in regular intervals.

    Methods
    -------
    mouse_handler(MouseEventArguments)
        Display the image ROI or intensity profile when clicking on the image.
    show_key(ValueChangeEventArguments, str)
        Display the key of the HDF dataset/group when a tree node is clicked.
    pick_file()
        Open a file picker dialog to select a file.
    display_hdf_tree(str)
        Display the HDF file structure as an interactive tree.
    disable_sliders()
        Disable and reset the sliders for 3D-data slicing.
    enable_ui_elements_3d_data()
        Enable UI elements specific to 3D-data display.
    enable_ui_elements_1d_2d_data()
        Enable UI elements specific to 1D/2D data display.
    reset(keep_display=False)
        Reset the UI elements to their initial states.
    reset_min_max()
        Reset the minimum and maximum sliders for image contrast.
    display_3d_data(data_obj)
        Display a slice of a 3D dataset as an image.
    display_1d_2d_data(data_obj, "plot")
        Display 1D/2D data as a plot or table.
    show_data()
        Display data from an HDF file based on the current GUI state.
    save_image()
        Save the currently displayed image to a file.
    save_data()
        Save the currently displayed 1D/2D data to a file.
    shutdown()
        Routine to close the app.
    """

    def __init__(self):
        super().__init__()
        self.select_file_button.on("click", self.pick_file)
        self.save_image_button.on("click", self.save_image)
        self.save_data_button.on("click", self.save_data)
        self.reset_button.on("click", self.reset_min_max)
        self.enable_zoom.on("click", self.__update_zoom_check_box)
        self.enable_profile.on("click", self.__update_profile_check_box)
        self.main_plot.on("click", self.mouse_handler)
        self.tab_one.on("click", self.__select_tab_one)
        self.tab_two.on("click", self.__select_tab_two)
        self.current_state, self.image, self.image_norm = None, None, None
        self.columns, self.rows = None, None
        self.current_slice, self.data_1d_2d = None, None
        self.timer = ui.timer(re.UPDATE_RATE, lambda: self.show_data())
        self.tab_one.on("click", self.__select_tab_one)
        self.tab_two.on("click", self.__select_tab_two)
        self.selected_tab = 1
        self.last_folder = ""
        self.fig, self.ax = None, None
        self.ver_line, self.hor_line, self.draw_roi = None, None, None

    def __select_tab_one(self):
        self.selected_tab = 1

    def __select_tab_two(self):
        self.selected_tab = 2

    def __update_zoom_check_box(self):
        if (self.enable_zoom.value is True
                and self.enable_profile.value is True):
            self.enable_profile.set_value(False)

    def __update_profile_check_box(self):
        if (self.enable_profile.value is True
                and self.enable_zoom.value is True):
            self.enable_zoom.set_value(False)

    def __get_xy(self, x, y, ax):
        try:
            xn, yn = ax.transData.inverted().transform((x, y))
            return xn, yn
        except Exception as e:
            return None, None

    def mouse_handler(self, e: events.MouseEventArguments):
        """
        Show the zoomed area around the mouse-clicked location or the
        intensity profile across the clicked location.
        """
        if self.image is not None and (
                self.enable_profile.value or self.enable_zoom.value):
            x, y = self.__get_xy(e.args['offsetX'], e.args['offsetY'], self.ax)
            if x is not None:
                (height, width) = self.image.shape
                x = int(x)
                y_max = self.ax.transData.inverted().transform(
                    (0, self.ax.get_ylim()[-1]))[-1]
                y = height - 1 - int(y) + int((y_max - height) / 2)
                if width > x >= 0 and height > y >= 0:
                    zp_fig = self.zoom_profile_plot.figure
                    zp_fig.clf()
                    zp_fig.set_dpi(self.dpi)
                    zp_ax = zp_fig.gca()
                    if self.draw_roi is not None:
                        self.draw_roi.set_visible(False)
                        self.draw_roi = None
                    if self.ver_line is not None:
                        self.ver_line.set_visible(False)
                        self.ver_line = None
                    if self.hor_line is not None:
                        self.hor_line.set_visible(False)
                        self.hor_line = None
                    if self.enable_profile.value:
                        if self.profile_list.value == "vertical":
                            self.ver_line = Line2D([x, x],
                                                   [-1, height + 1],
                                                   color=re.BOX_LINE_COLOR,
                                                   linewidth=\
                                                       re.BOX_LINE_WIDTH)
                            if self.ax is not None:
                                self.ax.add_line(self.ver_line)
                                list_data = self.image[:, x]
                                zp_ax.plot(list_data)
                                zp_ax.set_title(f"Profile at column: {x}")
                        else:
                            self.hor_line = Line2D([-1, width + 1], [y, y],
                                                   color=re.BOX_LINE_COLOR,
                                                   linewidth=\
                                                       re.BOX_LINE_WIDTH)
                            if self.ax is not None:
                                self.ax.add_line(self.hor_line)
                                list_data = self.image[y]
                                zp_ax.plot(list_data)
                                zp_ax.set_title(f"Profile at row: {y}")
                        zp_ax.set_xlabel("Index")
                        zp_ax.set_ylabel("Intensity")
                        zp_fig.tight_layout()
                        self.main_plot.update()
                    else:
                        if self.image_norm is not None:
                            val = self.zoom_list.value
                            zoom = int(val.replace("x", ""))
                            roi_img, x0, y0, size = \
                                util.get_image_roi(x, y, self.image_norm,
                                                   zoom=zoom)
                            self.draw_roi = patches.Rectangle(
                                (x0, y0), size, size,
                                linewidth=re.BOX_LINE_WIDTH,
                                edgecolor=re.BOX_LINE_COLOR,
                                facecolor='none')
                            self.ax.add_patch(self.draw_roi)
                            self.main_plot.update()
                            zp_ax.imshow(roi_img, cmap=self.cmap_list.value)
                            zp_fig.tight_layout()
                    self.zoom_profile_plot.update()

    def show_key(self, event: events.ValueChangeEventArguments, file_path):
        """
        Show key to a dataset/group of a hdf file when users click
        to a branch of the hdf tree.
        """
        hdf_key = event.value
        if hdf_key is not None:
            self.hdf_key_display.set_text(hdf_key)
        else:
            self.hdf_value_display.set_text("")
        if file_path is not None:
            self.file_path_display.set_text(file_path)

    async def pick_file(self) -> None:
        """To pick a file when click the button 'Select file' """
        config_data = util.load_config()
        if config_data is None:
            self.last_folder = ""
        else:
            try:
                self.last_folder = config_data["last_folder"]
            except KeyError:
                self.last_folder = ""
        if (self.last_folder == "") or (not os.path.exists(self.last_folder)):
            file_path = await FilePicker("~",
                                         allowed_extensions=re.INPUT_EXT)
        else:
            file_path = await FilePicker(self.last_folder,
                                         allowed_extensions=re.INPUT_EXT)
        if file_path:
            self.last_folder = os.path.dirname(file_path)
            config_data = {'last_folder': self.last_folder}
            util.save_config(config_data)
            self.display_hdf_tree(file_path)

    def display_hdf_tree(self, file_path):
        """Display interactive tree structure of a hdf file"""
        with self.tree_container:
            file_path = file_path.replace("\\", "/")
            self.file_path_display.set_text(file_path)
            self.hdf_key_display.set_text("")
            hdf_dic = util.hdf_tree_to_dict(file_path)
            if isinstance(hdf_dic, list):
                tree_display = ui.card()

                def close_file():
                    if isinstance(self.current_state, tuple):
                        if file_path == self.current_state[0]:
                            self.reset()
                    else:
                        self.reset()
                    self.tree_container.remove(tree_display)

                with tree_display.style("background-color: "
                                        + re.TREE_BGR_COLOR):
                    ui.tree(hdf_dic, label_key="id", node_key="label",
                            on_select=lambda e: self.show_key(e, file_path))
                    ui.button("Close file", on_click=lambda: close_file())
            else:
                if isinstance(hdf_dic, str):
                    ui.notify(hdf_dic)
                else:
                    ui.notify("Input must be hdf, hdf5, nxs, or h5 format!")

    def disable_sliders(self):
        """Disable and reset values of sliders"""
        self.main_slider.set_value(0)
        self.main_slider.disable()
        self.min_slider.set_value(0)
        self.min_slider.disable()
        self.max_slider.set_value(255)
        self.max_slider.disable()

    def enable_ui_elements_3d_data(self):
        """
        Enable UI-elements for displaying a slice from 3d data as an image and
        disable non-related UI-elements.
        """
        # Enable ui-components related to image show
        self.main_slider.enable()
        self.max_slider.enable()
        self.min_slider.enable()
        self.main_plot.set_visibility(True)
        self.axis_list.enable()
        self.cmap_list.enable()
        self.enable_zoom.enable()
        self.zoom_list.enable()
        self.enable_profile.enable()
        self.profile_list.enable()
        self.save_image_button.enable()
        self.histogram_plot.set_visibility(True)
        self.image_info_table.set_visibility(True)
        if self.enable_profile.value or self.enable_zoom.value:
            self.zoom_profile_plot.figure.clf()
            self.zoom_profile_plot.set_visibility(True)
        else:
            self.zoom_profile_plot.figure.clf()
            self.zoom_profile_plot.set_visibility(False)
            self.image_norm = None

        # Disable other ui-components
        self.main_table.set_visibility(False)
        self.display_type.disable()
        self.marker_list.disable()
        self.save_data_button.disable()
        self.data_1d_2d = None

    def enable_ui_elements_1d_2d_data(self):
        """
        Enable UI-elements for displaying 1d/2d data as a table or plot, and
        disable non-related UI-elements.
        """
        self.image, self.image_norm = None, None
        # Disable ui-components related to image show
        self.disable_sliders()
        self.axis_list.value = re.AXIS_LIST[0]
        self.cmap_list.value = re.CMAP_LIST[0]
        self.axis_list.disable()
        self.cmap_list.disable()
        self.enable_zoom.disable()
        self.zoom_list.disable()
        self.enable_profile.disable()
        self.profile_list.disable()
        self.save_image_button.disable()
        self.histogram_plot.set_visibility(False)
        self.image_info_table.set_visibility(False)
        self.zoom_profile_plot.set_visibility(False)

        # Enable other ui-components
        self.display_type.enable()
        self.marker_list.enable()
        self.save_data_button.enable()
        self.panel_tabs.set_value(self.tab_one)
        self.selected_tab = 1

    def reset(self, keep_display=False):
        """Reset status of UI-elements"""
        if not keep_display:
            self.hdf_key_display.set_text("")
            self.file_path_display.set_text("")
            self.hdf_value_display.set_text("")
        self.axis_list.value = re.AXIS_LIST[0]
        self.cmap_list.value = re.CMAP_LIST[0]
        self.display_type.value = re.DISPLAY_TYPE[0]
        self.marker_list.value = re.MARKER_LIST[0]
        self.axis_list.disable()
        self.cmap_list.disable()
        self.display_type.disable()
        self.marker_list.disable()
        self.enable_zoom.set_value(False)
        self.enable_zoom.disable()
        self.zoom_list.disable()
        self.enable_profile.set_value(False)
        self.enable_profile.disable()
        self.profile_list.disable()
        self.disable_sliders()
        self.rows, self.columns = None, None
        self.image, self.image_norm, self.data_1d_2d = None, None, None
        self.main_table.set_visibility(False)
        self.main_plot.set_visibility(True)
        self.zoom_profile_plot.set_visibility(False)
        self.save_image_button.disable()
        self.save_data_button.disable()
        self.histogram_plot.set_visibility(False)
        self.image_info_table.set_visibility(False)
        self.zoom_profile_plot.set_visibility(False)
        self.panel_tabs.set_value(self.tab_one)
        self.selected_tab = 1

    def reset_min_max(self):
        """Reset minimum and maximum values of sliders"""
        self.min_slider.set_value(0)
        self.max_slider.set_value(255)

    def display_3d_data(self, data_obj):
        """Display a slice of 3d array as an image"""
        self.enable_ui_elements_3d_data()
        (depth, height, width) = data_obj.shape
        current_max = self.main_slider._props["max"]
        if int(self.axis_list.value) == 2:
            max_val = width - 1
        elif int(self.axis_list.value) == 1:
            max_val = height - 1
        else:
            max_val = depth - 1
        if current_max != max_val:
            self.main_slider._props["max"] = max_val
            self.main_slider.update()
        d_pos = int(self.main_slider.value)
        if d_pos > max_val:
            self.main_slider.set_value(max_val)
            d_pos = max_val
        new_slice = (self.main_slider.value, self.axis_list.value,
                     self.file_path_display.text)
        if new_slice != self.current_slice or self.image is None:
            self.current_slice = new_slice
            if int(self.axis_list.value) == 2:
                if depth > 1000 and height > 1000:
                    ui.notify("Slicing along axis 2 is very time-consuming!")
                    self.axis_list.value = 0
                    self.main_slider.set_value(0)
                    self.image = data_obj[0]
                else:
                    self.image = data_obj[:, :, d_pos]
            elif int(self.axis_list.value) == 1:
                if depth > 1000 and width > 1000:
                    ui.notify("Slicing along axis 1 can take time !")
                self.image = data_obj[:, d_pos, :]
            else:
                self.image = data_obj[d_pos]
        min_val = int(self.min_slider.value)
        max_val = int(self.max_slider.value)
        if min_val > 0 or max_val < 255:
            if min_val >= max_val:
                min_val = np.clip(max_val - 1, 0, 254)
                self.min_slider.set_value(min_val)
            nmin, nmax = np.min(self.image), np.max(self.image)
            if nmax != nmin:
                self.image_norm = np.uint8(
                    255.0 * (self.image - nmin) / (nmax - nmin))
                self.image_norm = np.clip(self.image_norm, min_val, max_val)
            else:
                self.image_norm = np.zeros(self.image.shape)
        else:
            self.image_norm = np.copy(self.image)

        self.fig = self.main_plot.figure
        self.fig.clf()
        self.fig.set_dpi(self.dpi)
        self.ax = self.fig.gca()
        self.ax.imshow(self.image_norm, cmap=self.cmap_list.value)
        self.fig.tight_layout()
        self.main_plot.update()

        if self.selected_tab == 2:
            rows, columns = util.format_statistical_info(self.image)
            self.image_info_table.rows[:] = rows
            self.image_info_table.columns[:] = columns
            self.image_info_table.update()
            with self.histogram_plot:
                plt.clf()
                flat_data = self.image.ravel()
                num_bins = min(255, len(flat_data))
                hist, bin_edges = np.histogram(flat_data, bins=num_bins)
                plt.hist(bin_edges[:-1], bins=bin_edges, weights=hist,
                         color='skyblue', edgecolor='black', alpha=0.65,
                         label=f"Num bins: {num_bins}")
                plt.title("Histogram")
                plt.xlabel("Grayscale")
                plt.ylabel("Frequency")
                plt.legend()
                self.histogram_plot.update()

    def display_1d_2d_data(self, data_obj, disp_type="plot"):
        """Display 1d/2d array as a table or plot"""
        self.enable_ui_elements_1d_2d_data()
        self.data_1d_2d = data_obj[:]
        if disp_type == "table":
            self.main_plot.set_visibility(False)
            self.main_table.set_visibility(True)
            rows, columns = util.format_table_from_array(data_obj[:])
            self.main_table.rows[:] = rows
            self.main_table.columns[:] = columns
            self.main_table.update()
        else:
            self.main_plot.set_visibility(True)
            self.main_table.set_visibility(False)
            x, y = None, None
            img = False
            if len(data_obj.shape) == 2:
                (height, width) = data_obj.shape
                if height == 2:
                    x, y = np.asarray(data_obj[0]), np.asarray(data_obj[1])
                elif width == 2:
                    x = np.asarray(data_obj[:, 0])
                    y = np.asarray(data_obj[:, 1])
                else:
                    img = True
            else:
                size = len(data_obj)
                x, y = np.arange(size), np.asarray(data_obj[:])
            if x is not None:
                title = self.hdf_key_display.text.split("/")[-1]
                fig = self.main_plot.figure
                fig.clf()
                fig.set_dpi(self.dpi)
                ax = fig.gca()
                ax.set_title(title.capitalize())
                ax.plot(x, y, marker=self.marker_list.value,
                        color=re.PLOT_COLOR)
                fig.tight_layout()
                self.main_plot.update()
            if img:
                fig = self.main_plot.figure
                fig.clf()
                fig.set_dpi(self.dpi)
                ax = fig.gca()
                ax.imshow(data_obj[:], cmap=self.cmap_list.value,
                          aspect="auto")
                fig.tight_layout()
                self.main_plot.update()

    def __clear_plot(self):
        self.main_plot.figure.clf()
        self.main_plot.update()
        self.zoom_profile_plot.figure.clf()
        self.zoom_profile_plot.update()
        with self.histogram_plot:
            plt.clf()
            self.histogram_plot.update()

    def show_data(self):
        """Display data getting from a hdf file"""
        file_path1 = self.file_path_display.text
        hdf_key1 = self.hdf_key_display.text
        if (file_path1 != "") and (hdf_key1 != "") and (hdf_key1 is not None):
            new_state = (file_path1, hdf_key1, self.main_slider.value,
                         self.hdf_value_display.text, self.axis_list.value,
                         self.cmap_list.value, self.display_type.value,
                         self.marker_list.value, self.min_slider.value,
                         self.max_slider.value, self.selected_tab,
                         self.enable_zoom.value, self.enable_profile.value)
            if new_state != self.current_state:
                self.current_state = new_state
                try:
                    (data_type, value) = util.get_hdf_data(file_path1,
                                                           hdf_key1)
                    if (data_type == "string" or data_type == "number"
                            or data_type == "boolean"):
                        self.hdf_value_display.set_text(str(value))
                        self.__clear_plot()
                        self.reset(keep_display=True)
                    elif data_type == "array":
                        self.hdf_value_display.set_text("Array shape: "
                                                        "" + str(value))
                        hdf_obj = h5py.File(file_path1, "r")
                        dim = len(value)
                        if dim == 3:
                            self.display_3d_data(hdf_obj[hdf_key1])
                        elif dim < 3:
                            self.display_1d_2d_data(
                                hdf_obj[hdf_key1],
                                disp_type=self.display_type.value)
                        else:
                            ui.notify("Can't display {}-d array!".format(dim))
                            self.__clear_plot()
                            self.reset(keep_display=True)
                        hdf_obj.close()
                    else:
                        self.hdf_value_display.set_text(data_type)
                        self.__clear_plot()
                        self.reset(keep_display=True)
                except Exception as error:
                    self.reset(keep_display=True)
                    _, broken_link, msg = util.check_external_link(file_path1,
                                                                   hdf_key1)
                    if broken_link:
                        ui.notify(msg)
                    else:
                        _, ext_compressed, msg = util.check_compressed_dataset(
                            file_path1, hdf_key1)
                        if ext_compressed:
                            ui.notify(msg)
                        else:
                            ui.notify("Error: {}".format(error))
                            ui.notify("Dataset may be an external link and the"
                                      " target file is not accessible (moved, "
                                      "deleted, or corrupted) !!!")
        else:
            self.hdf_value_display.set_text("")
            self.__clear_plot()
            self.reset(keep_display=True)

    async def save_image(self) -> None:
        """To save a slice to file when click 'Save image' """
        if (self.last_folder == "") or (not os.path.exists(self.last_folder)):
            file_path = await FileSaver("~", title="File name (ext: .tif, "
                                                   ".jpg, .png, or .csv)")
        else:
            file_path = await FileSaver(self.last_folder,
                                        title="File name (ext: .tif, "
                                              ".jpg, .png, or .csv)")
        if file_path and self.image is not None:
            file_ext = os.path.splitext(file_path)[-1]
            if (file_ext != ".tif" and file_ext != ".jpg"
                    and file_ext != ".png" and file_ext != ".csv"):
                ui.notify("Please use .tif, .jpg, .png, or .csv as "
                          "file extension!")
            else:
                check = os.path.isfile(file_path)
                if file_ext == ".csv":
                    error = util.save_table(file_path, self.image)
                else:
                    error = util.save_image(file_path, self.image)
                if error is not None:
                    ui.notify(error)
                else:
                    if check:
                        ui.notify(
                            "File {} is overwritten".format(file_path))
                    else:
                        ui.notify("File is saved at: {}".format(file_path))

    async def save_data(self) -> None:
        """To save data to file when click the button 'Save data' """
        if (self.last_folder == "") or (not os.path.exists(self.last_folder)):
            file_path = await FileSaver("~", title="File name (ext: .csv)")
        else:
            file_path = await FileSaver(self.last_folder,
                                        title="File name (ext: .csv)")
        if file_path and self.data_1d_2d is not None:
            file_ext = os.path.splitext(file_path)[-1]
            if file_ext == "":
                file_ext = ".csv"
                file_path = file_path + file_ext
            if file_ext != ".csv":
                ui.notify("Please use .csv as file extension!")
            else:
                check = os.path.isfile(file_path)
                error = util.save_table(file_path, self.data_1d_2d)
                if error is not None:
                    ui.notify(error)
                else:
                    if check:
                        ui.notify(
                            "File {} is overwritten".format(file_path))
                    else:
                        ui.notify(
                            "File is saved at: {}".format(file_path))

    def shutdown(self):
        """Routine to close the app"""
        self.timer.cancel()
        ui.notify("The server has been stopped. You can close this tab!")



"""
Module for utility methods:

    -   Get height and with of a screen.
    -   Convert hdf tree to a nested dictionary.
    -   Get data-type and value from a dataset in a hdf file.
    -   Convert 1d/2d array to a table format.
    -   Save 2d array to an image.
    -   Save 1d/2d array to a csv file.
    -   Save/get path of the last opened folder
"""

import os
import json
import platform
import csv
import tkinter as tk
import h5py
import hdf5plugin
import numpy as np
from PIL import Image


def get_height_width_screen():
    """
    Get the height and width of the current screen.

    Returns
    -------
    tuple
       A tuple of (screen height, screen width).
    """
    root = tk.Tk()
    screen_width = root.winfo_screenwidth()
    screen_height = root.winfo_screenheight()
    dpi = root.winfo_fpixels('1i')
    root.destroy()
    return screen_height, screen_width, dpi


def __recurse(parent_path, name, obj):
    """
    Supplementary function for recursive traversal of HDF5 file structure.

    Parameters
    ----------
    parent_path : str
        The path of the parent group.
    name : str
        The name of the current group or dataset.
    obj : h5py-object
        The current HDF5 group or dataset.

    Returns
    -------
    list
        A list of dictionaries describing the HDF5 groups and datasets.
    """
    current_path = f"{parent_path}/{name}" if parent_path else name
    if isinstance(obj, h5py._hl.dataset.Dataset):
        return [{"id": name, "label": current_path}]
    elif isinstance(obj, h5py._hl.group.Group):
        children = []
        for key in obj.keys():
            try:
                child_obj = obj[key]
                child_items = __recurse(current_path, key, child_obj)
                children.extend(child_items)
            except Exception:
                children.append({"id": key, "label": f"{current_path}/{key}"})
        return [{"id": name, "label": current_path, "children": children}]


def hdf_tree_to_dict(hdf_file):
    """
    Convert an HDF5 file structure to a nested dictionary.

    Parameters
    ----------
    hdf_file : str
        Path to the HDF5 file.

    Returns
    -------
    dict or str
        A nested dictionary representing the structure of the HDF5 file,
        or a string describing an error if one occurs.
    """
    try:
        hdf_obj1 = h5py.File(hdf_file, 'r')
        result = __recurse("", "", hdf_obj1)
        file_name = os.path.basename(hdf_file)
        result[0]['id'] = file_name
        result[0]['label'] = "/"
        hdf_obj1.close()
        return result
    except Exception as error:
        return str(error)


def get_hdf_data(file_path, dataset_path):
    """
    Get data type and value from a specified dataset in an HDF5 file.

    Parameters
    ----------
    file_path : str
        Path to the HDF5 file.
    dataset_path : str
        Path to the dataset within the HDF5 file.

    Returns
    -------
    tuple
        A tuple containing the data type and the value of the dataset.
    """
    with h5py.File(file_path, 'r') as file:
        if dataset_path not in file:
            return "not path", None
        try:
            item = file[dataset_path]
            if isinstance(item, h5py.Group):
                return "group", None
            data_type, value = "unknown", None
            # Check the type and shape of a dataset
            if item.dtype.kind == 'S':  # Fixed-length bytes
                data = item[()]
                if item.size == 1:  # Single string or byte
                    if isinstance(data, bytes):
                        data_type, value = "string", data.decode('utf-8')
                    elif isinstance(data.flat[0], bytes):
                        data_type, value = "string", data.flat[0].decode(
                            'utf-8')
                else:
                    data_type, value = "array", [d.decode('utf-8') for d in
                                                 data]
            elif item.dtype.kind == 'U':  # Fixed-length Unicode
                data = item[()]
                if item.size == 1:  # Single string
                    data_type, value = "string", data
                else:
                    data_type, value = "array", list(data)
            elif h5py.check_dtype(vlen=item.dtype) in [str, bytes]:
                data = item[()]
                if isinstance(data, (str, bytes)):
                    data_type, value = "string", data if isinstance(data, str)\
                        else data.decode('utf-8')
                else:
                    joined_data = ''.join(
                        [d if isinstance(d, str) else d.decode('utf-8')
                         for d in data])
                    data_type, value = "string", joined_data
            elif item.dtype.kind in ['i', 'f', 'u']:
                if item.shape == () or item.size == 1:
                    data_type, value = "number", item[()]
                else:
                    data_type, value = "array", item.shape
            elif item.dtype.kind == 'b':  # Boolean type
                data_type, value = "boolean", int(item[()])
            return data_type, value
        except Exception as error:
            return str(error), None


def check_external_link(file_path, dataset_path):
    """
    Check if the dataset at a specified path is an external link and
    whether it's broken.

    Parameters
    ----------
    file_path : str
        Path to the HDF5 file.
    dataset_path : str
        Path to the dataset within the HDF5 file.

    Returns
    -------
    tuple
        A tuple containing a boolean indicating if it's an external link,
        a boolean indicating if the link is broken, and a message string.
    """
    ext_link = False
    broken = False
    msg = ""
    with h5py.File(file_path, 'r') as file:
        if dataset_path in file:
            # Check for external link
            link = file.get(dataset_path, getlink=True)
            if isinstance(link, h5py.ExternalLink):
                ext_link = True
                try:
                    file[dataset_path]
                    msg = "Dataset is an external link"
                except Exception as error:
                    msg = "Dataset is an external link but failed " \
                          "to link. Error: " + str(error)
                    broken = True
        else:
            msg = "Dataset path not found in the file."
    return ext_link, broken, msg


def check_compressed_dataset(file_path, dataset_path):
    """
    Check if a dataset is compressed, including checking for external
    compression filters.

    Parameters
    ----------
    file_path : str
        Path to the HDF5 file.
    dataset_path : str
        Path to the dataset within the HDF5 file.

    Returns
    -------
    tuple
        A tuple containing two booleans indicating if the dataset is compressed
        and if it uses external compression filters, and a message string.
    """
    compressed = False
    ext_compressed = False
    msg = ""
    with h5py.File(file_path, 'r') as file:
        if dataset_path in file:
            dataset = file[dataset_path]
            compression = dataset.compression
            # Check for standard compression
            if compression:
                compressed = True
                msg = "Dataset is compressed using standard method: " \
                      "{}".format(compression)
            # Check for external filters
            plist = dataset.id.get_create_plist()
            n_filters = plist.get_nfilters()
            for i in range(n_filters):
                filter_info = plist.get_filter(i)
                if filter_info:
                    msg = "Dataset is compressed using external plugin:" \
                          " {}".format(filter_info[-1])
                    ext_compressed = True
        else:
            msg = "Dataset path not found in the file."
    return compressed, ext_compressed, msg


def format_table_from_array(data):
    """
    Format a NumPy array into a structure suitable for displaying as a table
    using NiceGUI framework.

    Parameters
    ----------
    data : ndarray
        The NumPy array to format.

    Returns
    -------
    tuple
        A tuple containing the rows and columns formatted for the table.
    """
    if len(data.shape) == 1:
        data = np.expand_dims(np.asarray(data), 1)
    (height, width) = data.shape[:2]
    if width > height:
        data = np.transpose(data)
        (height, width) = data.shape[:2]
    if height > 1000 and width > 1000:
        rows, columns = None, None
    else:
        fm_data = np.insert(data, 0, np.arange(height), axis=1)
        columns = [{"name": "Column " + str(j), "label": "Column " + str(j),
                    "field": "Column " + str(j)} for j in range(width)]
        columns.insert(0,
                       {"name": "Index", "label": "Index", "field": "Index"})
        rows = []
        for i in range(height):
            dict_tmp = {}
            for j in range(width + 1):
                dict_tmp[columns[j]["name"]] = fm_data[i][j]
            rows.append(dict_tmp)
    return rows, columns


def format_statistical_info(image):
    """
    Get statistical information of a 2d array and format the output as a
    Nicegui table object.

    Parameters
    ----------
    image : ndarray
        NumPy array to format.

    Returns
    -------
    tuple
        A tuple containing the rows and columns formatted for the table.
    """
    data_type = image.dtype.name
    min_val = np.min(image)
    max_val = np.max(image)
    mean_val = np.mean(image)
    std_val = np.std(image)
    columns = [{"name": "information", "label": "Information",
                "field": "information"},
               {"name": "value", "label": "Value", "field": "value"}]
    rows = [{"name": "Minimum", "value": min_val},
            {"name": "Maximum", "value": max_val},
            {"name": "Mean", "value": mean_val},
            {"name": "Standard deviation", "value": std_val},
            {"name": "Data type", "value": data_type}]
    return rows, columns


def save_image(file_path, mat):
    """
    Save a 2D array as an image file.

    Parameters
    ----------
    file_path : str
        Path where the image will be saved.
    mat : ndarray
        2D array to be saved as an image.

    Returns
    -------
    None or str
        Returns None if successful, or a string message if an error occurs.
    """
    file_ext = os.path.splitext(file_path)[-1]
    if not ((file_ext == ".tif") or (file_ext == ".tiff")):
        mat = np.uint8(
            255.0 * (mat - np.min(mat)) / (np.max(mat) - np.min(mat)))
    else:
        if mat.dtype != np.float32:
            mat = mat.astype(np.float32)
    image = Image.fromarray(mat)
    try:
        image.save(file_path)
    except Exception as error:
        return str(error)


def save_table(file_path, data):
    """
    Save a 1D or 2D NumPy array to a CSV file.

    Parameters
    ----------
    file_path : str
        Path where the CSV file will be saved.
    data : ndarray
        The 1D or 2D array to be saved.

    Returns
    -------
    None or str
        Returns None if successful, or a string message if an error occurs.
    """
    try:
        data = np.asarray(data)
        with open(file_path, 'w', newline='') as file:
            writer = csv.writer(file)
            if data.ndim == 1:
                for item in data:
                    writer.writerow([item])
            elif data.ndim == 2:
                if data.shape[0] * data.shape[1] < 4000000:
                    writer.writerows(data)
                else:
                    return "Array has more than 4,000,000 elements. " \
                           "Operation not performed."
            else:
                return "Data must be a 1D or 2D array"
    except Exception as error:
        return str(error)


def get_config_path():
    """
    Get path to save a config file depending on the OS system.
    """
    home = os.path.expanduser("~")
    if platform.system() == "Windows":
        return os.path.join(home, 'AppData', 'Roaming', 'Broh5',
                            'broh5_config.json')
    elif platform.system() == "Darwin":
        return os.path.join(home, 'Library', 'Application Support', 'Broh5',
                            'broh5_config.json')
    else:
        return os.path.join(home, '.broh5', 'broh5_config.json')


def save_config(data):
    """
    Save data (dictionary) to the config file (json format).
    """
    config_path = get_config_path()
    os.makedirs(os.path.dirname(config_path), exist_ok=True)
    with open(config_path, 'w') as f:
        json.dump(data, f)


def load_config():
    """
    Load the config file.
    """
    config_path = get_config_path()
    try:
        with open(config_path, 'r') as f:
            return json.load(f)
    except FileNotFoundError:
        return None


def get_image_roi(x, y, mat, zoom=2):
    """
    Get the square ROI (Region of Interest) of an image given a zoom
    value and the region center.

    Parameters
    ----------
    x : int
        x-center of the squared ROI
    y : int
        y-center of the squared ROI
    mat : ndarray
        2d array.
    zoom : float
        Zoom-in value of the ROI.

    Returns
    -------
    roi : ndarray
        ROI of the image
    x_start : int
        x-start of the ROI
    y_start : int
        y_start of the ROI
    size : int
        Size of the ROI.
    """
    (height, width) = mat.shape
    min_size = min(height, width)
    rad = np.clip(int(0.5 * min_size // zoom), 1, min_size // 2 - 1)
    size = 2 * rad
    x_start = np.clip(x - rad, 0, width - size)
    y_start = np.clip(y - rad, 0, height - size)
    x_stop = x_start + size
    y_stop = y_start + size
    roi = mat[y_start:y_stop, x_start:x_stop]
    return roi, x_start, y_start, size



import pathlib
import setuptools

dependencies = [
    "numpy>=1.23",
    "h5py>3.7",
    "hdf5plugin",
    "pillow",
    "matplotlib",
    "nicegui>=3.0"
]

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setuptools.setup(
    name="broh5",
    version="1.4.2",
    author="Nghia Vo",
    author_email="nvo@bnl.gov",
    description='Browser-based GUI HDF Viewer in Python',
    long_description=README,
    long_description_content_type="text/markdown",
    keywords=['HDF Viewer', 'NXS Viewer', 'Browser-based GUI'],
    url="https://github.com/algotom/broh5",
    download_url="https://github.com/algotom/broh5.git",
    license="Apache 2.0",
    platforms="Any",
    packages=setuptools.find_packages(include=["broh5", "broh5.*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering"
    ],
    install_requires=dependencies,
    entry_points={'console_scripts': ['broh5 = broh5.main:main']},
    python_requires='>=3.9',
)
